


# https://www.ocregister.com/2018/12/17/big-surf-possible-flooding-and-erosion-as-massive-swell-hits-the-coast-this-week/

# field definitions
# https://www.ndbc.noaa.gov/measdes.shtml

# full data set
# https://www.ndbc.noaa.gov/view_text_file.php?filename=46222h2018.txt.gz&dir=data/historical/stdmet/


# Fitting Many Dimensions into One: The Promise of Hierarchical Indices for Data Beyond Two Dimensions

# -----------------------------------------------
import typing as tp
import os
import pickle
from urllib import request
from typing import NamedTuple
import functools

import numpy as np
import pandas as pd

import static_frame as sf
from static_frame.performance.perf_test import PerfTest


class _PerfTestPanel(PerfTest):

    NUMBER = 500
    FUNCTION_NAMES = ('dict_df', 'panel_float', 'np', 'xarray_da', 'pd_mi_2D', 'sf_ih_2D', 'pd_mi_1D', 'sf_ih_1D')



#-------------------------------------------------------------------------------
#

class Buoy(NamedTuple):
    station_id: int
    name: str

BUOYS = (
    Buoy(46222, 'San Pedro'),
    Buoy(46253, 'San Pedro South'),
    Buoy(46221, 'Santa Monica Bay'),
)

def cache_buoy(prefix, active=True):
    def decorator(func):
        def wrapper(cls, buoy, year):
            fp = f'/tmp/{prefix}-{buoy.station_id}-{year}.p'
            load_source = True
            if active and os.path.exists(fp):
                with open(fp, 'rb') as f:
                    try:
                        post = pickle.load(f)
                        load_source = False
                    except ModuleNotFoundError:
                        pass
            if load_source:
                post = func(cls, buoy, year)
                with open(fp, 'wb') as f:
                    pickle.dump(post, f)
            return post
        return wrapper
    return decorator


class BuoyLoader:

    FIELD_DATETIME = 'datetime'
    FIELD_STATION_ID = 'station_id'
    FIELD_WAVE_HEIGHT = 'WVHT'
    FIELD_WAVE_PERIOD = 'DPD'	# Dominant wave period
    FIELD_WAVE_DIRECTION = 'MWD'

    COMPASS = ('N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE', 'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW')
    URL_TEMPLATE = 'https://www.ndbc.noaa.gov/view_text_file.php?filename={station_id}h{year}.txt.gz&dir=data/historical/stdmet/'

    @classmethod
    def buoy_record(cls,
            line: str,
            line_number: int,
            station_id: int
            ) -> tp.Sequence[str]:
        timestamp = []

        def gen() -> tp.Iterator[tp.Union[int, str]]:
            yield cls.FIELD_STATION_ID if not line_number else station_id
            cell_pos = -1 # increment before usage
            for cell in line.split(' '):
                cell = cell.strip()
                if cell:
                    cell_pos += 1
                    if cell_pos < 5:
                        timestamp.append(cell)
                    elif cell_pos == 5:
                        yield cls.FIELD_DATETIME if not line_number else '{}-{}-{}T{}:{}'.format(*timestamp)
                        yield cell
                    else:
                        yield cell

        return tuple(gen())

    @classmethod
    def buoy_to_records(cls,
            buoy: Buoy,
            year: int,
            ) -> tp.Iterator[tp.Sequence[str]]:

        url = cls.URL_TEMPLATE.format(station_id=buoy.station_id, year=year)

        with request.urlopen(url) as response:
            raw = response.read().decode('utf-8')

        line_pos = -1 # increment before usage to allow skipped lines
        for line in raw.split('\n'):
            if not line.strip():
                continue
            line_pos += 1
            yield cls.buoy_record(line, line_pos, station_id=buoy.station_id)


    @classmethod
    def degree_to_compass(cls, degrees: np.array):
        indices = np.floor(((degrees + 12.25) % 360) / 22.5).astype(int)
        return np.array([cls.COMPASS[i] for i in indices])

    @classmethod
    @cache_buoy('sf')
    def buoy_to_sf(cls, buoy: Buoy, year: int) -> sf.Frame:
        '''
        Return a simple Frame presentation without an index.
        '''

        records = cls.buoy_to_records(buoy, year=year)
        columns = next(records)
        units = next(records)

        # can pass dtypes here, but doing it below consolidates blocks
        dtypes = {
                cls.FIELD_WAVE_HEIGHT: float,
                cls.FIELD_WAVE_PERIOD: float,
                cls.FIELD_DATETIME: np.datetime64}
        f = sf.Frame.from_records(records, columns=columns, dtypes=dtypes)

        direction = cls.degree_to_compass(f[cls.FIELD_WAVE_DIRECTION].astype(int).values)

        f = f[[cls.FIELD_STATION_ID,
                cls.FIELD_DATETIME,
                cls.FIELD_WAVE_HEIGHT,
                cls.FIELD_WAVE_PERIOD]].to_frame_go()

        f[cls.FIELD_WAVE_DIRECTION] = direction

        return f.to_frame()

    @classmethod
    @cache_buoy('df')
    def buoy_to_pd(cls, buoy: Buoy, year: int):

        records = cls.buoy_to_records(buoy, year=year)
        columns = next(records)
        units = next(records)

        df = pd.DataFrame.from_records(records, columns=columns)

        direction = cls.degree_to_compass(df[cls.FIELD_WAVE_DIRECTION].astype(int).values)

        df = df[[cls.FIELD_STATION_ID,
                cls.FIELD_DATETIME,
                cls.FIELD_WAVE_HEIGHT,
                cls.FIELD_WAVE_PERIOD]]

        df[cls.FIELD_WAVE_DIRECTION] = direction

        return df.astype({
                cls.FIELD_WAVE_HEIGHT: float,
                cls.FIELD_WAVE_PERIOD: float,
                cls.FIELD_DATETIME: np.datetime64})


#-------------------------------------------------------------------------------


class BuoySingleYear2D:

    @staticmethod
    def to_sf(year: int = 2018) -> sf.Frame:

        frames = []
        for buoy in BUOYS:
            f = BuoyLoader.buoy_to_sf(buoy, year)
            frames.append(f)

        f = sf.Frame.from_concat(frames,
                index=sf.IndexAutoFactory,
                )
        f = f.set_index_hierarchy(('station_id', 'datetime'),
                index_constructors=(sf.Index, sf.IndexMinute),
                drop=True)
        return f

    @staticmethod
    def to_pd(year: int = 2018):

        dfs = []
        for buoy in BUOYS:
            f = BuoyLoader.buoy_to_pd(buoy, year)
            dfs.append(f)

        df = pd.concat(dfs)

        return df.set_index(['station_id', 'datetime']).sort_index()

        # this sets to datetime, but does not a
    @staticmethod
    def to_pd_dict(year: int = 2018):

        dfs = {}
        for buoy in BUOYS:
            df = BuoyLoader.buoy_to_pd(buoy, year)
            df = df.set_index('datetime')[['DPD', 'WVHT', 'MWD']]
            dfs[buoy.station_id] = df

        return dfs


    @staticmethod
    def to_pd_panel_obj(year: int = 2018):
        dfs = {}
        for buoy in BUOYS:
            df = BuoyLoader.buoy_to_pd(buoy, year)
            df = df.set_index('datetime')[['DPD', 'WVHT', 'MWD']]
            dfs[buoy.station_id] = df

        return pd.Panel(dfs)


    @staticmethod
    def to_pd_panel_float(year: int = 2018):
        dfs = {}
        for buoy in BUOYS:
            df = BuoyLoader.buoy_to_pd(buoy, year)
            # remove station id, set datetime as indexnns
            # df = df.set_index('datetime')[['DPD', 'WVHT']]
            df = df.set_index('datetime')[['DPD', 'WVHT']]
            dfs[buoy.station_id] = df

        return pd.Panel(dfs)


    @classmethod
    def to_xarray_float(cls, year: int = 2018):
        # this is an object typed DataArray
        return cls.to_pd_panel_float(year=year).to_xarray()


    @classmethod
    def to_np(cls, year: int = 2018):
        arrays = []

        frames = [BuoyLoader.buoy_to_sf(buoy, year) for buoy in BUOYS]

        # NOTE: take union and inssert NaNs
        date_intersect = functools.reduce(lambda x, y: x | y, (set(f['datetime'].values) for f in frames))
        date_intersect = sorted(date_intersect)

        frames_aligned = []
        for f in frames:
            # frames_aligned.append(f.loc[f['datetime'].isin(date_intersect)])
            f = f.set_index('datetime', drop=True)
            f = f.reindex(date_intersect)
            f = f.unset_index()
            f = f.relabel(columns={'__index0__':'datetime'})
            frames_aligned.append(f)


        station_ids = {}
        for idx, f in enumerate(frames_aligned):
            datetime = {d: x for x, d in enumerate(f['datetime'].values)} # store the last one
            station_id = int(f.loc[0, 'station_id']) # new index counts from zero
            arrays.append(f[['DPD', 'WVHT']].values)
            station_ids[station_id] = idx

        indices = {'station_id': station_ids, 'datetime':  datetime, 'attr': {'DPD':0, 'WVHT':1}}
        # import ipdb; ipdb.set_trace()
        return np.array(arrays), indices


    #---------------------------------------------------------------------------

    @classmethod
    def process_sf(cls) -> None:

        fsf = cls.to_sf()

        #-----------------------------------------------------------------------
        # getting values out

        # iterating parts of the index
        part = tuple(fsf.index.iter_label(0))

        # getting all the values at an array
        part = fsf.index.values_at_depth(0)

        # convert to a frame
        part = fsf.index.to_frame()

        # rehiearch (make dates outer, station id inner); note that changing the hiearchy order forces a reordering
        part = fsf.rehierarch((1, 0))

        # to tuples
        part = fsf.relabel_flat(index=True)

        # adding a level
        part = fsf.relabel_level_add(index='A')

        # dropping a level
        #         ipdb> fsf.relabel_level_drop(index=1)
        # *** static_frame.core.exception.ErrorInitIndex: labels (50350) have non-unique values (17034)
        # ipdb>

        #-----------------------------------------------------------------------
        # show different types of selection

        # getting a sample from a partial match
        part = fsf.loc[sf.HLoc[:, '2018-12-18T07'], 'DPD']

        # select based on partial time
        post1 = fsf.loc[sf.HLoc[:, '2018-12-18T20']]

        # getting a slice
        part = fsf.loc[sf.HLoc[:, '2018-12-18T07':'2018-12-18T08'], 'DPD']

        # dsicrete selection (this works on Pandas)
        part = fsf.loc[sf.HLoc[:, ['2018-12-18T20:00', '2018-12-18T20:30']], 'DPD']

        # can show iloc
        part = fsf.loc[sf.HLoc[:, ['2018-12-18T20:00', '2018-12-18T20:30']], sf.ILoc[-1]]



        # show getting labels with iter_label (unique values)
        # show converting to a Frame

        #-----------------------------------------------------------------------
        # doing some analysis

        # find max for givne day
        fsf.loc[sf.HLoc[:, '2018-12-18']].max()


        max_dpd = [fsf.loc[sf.HLoc[station_id], 'DPD'].loc_max() for station_id in fsf.index.iter_label(0)]
        max_wvht = [fsf.loc[sf.HLoc[station_id], 'WVHT'].loc_max() for station_id in fsf.index.iter_label(0)]

        # get the peaks of the two fields, but this does not get us to the date
        peaks = fsf.loc[fsf.index.isin(max_dpd + max_wvht)]

        # use 2 to get 1.731622836825856
        threshold_wvht = fsf.loc[:, 'WVHT'].mean() + (fsf.loc[:, 'WVHT'].std() * 2)

        # use 1 to get 15.889409302831822
        threshold_dpd = fsf.loc[:, 'DPD'].mean() + fsf.loc[:, 'DPD'].std()

        # this isolates the relevant days; but does not get 46253
        # 2 and 18 gets all
        targets = fsf.loc[(fsf.loc[:, 'WVHT'] > threshold_wvht) & (fsf.loc[:, 'DPD'] > threshold_dpd)]

        targets = targets.to_frame_go()
        targets['date'] = [d.date() for d in targets.index.values_at_depth(1)]


        targets['station_id'] = targets.index.values_at_depth(0)
        targets.iter_group(['date', 'station_id']).apply(len)

        # targets.iter_group('date').apply(lambda x: len(x))
        peaks_per_day = targets.iter_group('date').apply(len)
        print(peaks_per_day)

        def gen():
            for date, date_frame in targets.iter_group_items('date'):
                for station_id, station in date_frame.iter_group_labels_items(0):
                    yield date, station_id, len(station)

        post = sf.Frame.from_records(gen(), columns=('date', 'station_id', 'count'))
        post = post.set_index_hierarchy(('date', 'station_id'),
                drop=True,
                index_constructors=(sf.IndexDate, sf.Index))

        print(post)


#-------------------------------------------------------------------------------
class BuoySingleYear1D:
    '''
    Fit all the data into a Series
    '''

    FIELD_ATTR = 'attr'

    @staticmethod
    def to_sf(year: int = 2018) -> sf.Frame:

        labels = []
        values = []

        for buoy in BUOYS:
            f = BuoyLoader.buoy_to_sf(buoy, year)
            for row in f.iter_series(1):
                for attr in (
                        BuoyLoader.FIELD_WAVE_HEIGHT,
                        BuoyLoader.FIELD_WAVE_PERIOD):
                    label = (row[BuoyLoader.FIELD_STATION_ID],
                            row[BuoyLoader.FIELD_DATETIME],
                            attr)
                    labels.append(label)
                    values.append(row[attr])

        index = sf.IndexHierarchy.from_labels(labels,
                index_constructors=(sf.Index, sf.IndexMinute, sf.Index))

        return sf.Series(values, index=index)


    @staticmethod
    def to_pd_obj(year: int = 2018) -> sf.Frame:

        labels = []
        values = []

        for buoy in BUOYS:
            df = BuoyLoader.buoy_to_pd(buoy, year)
            for _, row in df.iterrows():
                for attr in (
                        BuoyLoader.FIELD_WAVE_HEIGHT,
                        BuoyLoader.FIELD_WAVE_PERIOD,
                        BuoyLoader.FIELD_WAVE_DIRECTION,
                        ):
                    label = (row[BuoyLoader.FIELD_STATION_ID],
                            row[BuoyLoader.FIELD_DATETIME],
                            attr)
                    labels.append(label)
                    values.append(row[attr])

        # display of this index is terrible
        index = pd.MultiIndex.from_tuples(labels)
        return pd.Series(values, index=index).sort_index()



    @staticmethod
    def to_pd_float(year: int = 2018) -> sf.Frame:

        labels = []
        values = []

        for buoy in BUOYS:
            df = BuoyLoader.buoy_to_pd(buoy, year)
            for _, row in df.iterrows():
                for attr in (
                        BuoyLoader.FIELD_WAVE_HEIGHT,
                        BuoyLoader.FIELD_WAVE_PERIOD):
                    label = (row[BuoyLoader.FIELD_STATION_ID],
                            row[BuoyLoader.FIELD_DATETIME],
                            attr)
                    labels.append(label)
                    values.append(row[attr])

        # display of this index is terrible
        index = pd.MultiIndex.from_tuples(labels)
        return pd.Series(values, index=index).sort_index()


#-------------------------------------------------------------------------------
# performance tests

class SampleData:

    _store: tp.Dict[str, tp.Any] = {}


    @classmethod
    def create(cls) -> None:
        # fsf = BuoyLoader.buoy_to_sf(BUOYS[0], 2018)
        # fpd = BuoyLoader.buoy_to_pd(BUOYS[0], 2018)


        # cls._store['array_datetime'] = fsf['datetime'].values
        # cls._store['array_station_id'] = tuple(b.station_id for b in BUOYS)
        # cls._store['array_attr'] = ('WVHT', 'DPD', 'MWD')

        # cls._store['sf_index_datetime'] = sf.IndexMinute(cls.get('array_datetime'))
        # cls._store['sf_index_station_id'] = sf.Index(cls.get('array_station_id'))
        # cls._store['sf_index_attr'] = sf.Index(cls.get('array_attr'))

        # cls._store['pd_index_datetime'] = pd.Index(cls.get('array_datetime'))
        # cls._store['pd_index_station_id'] = pd.Index(cls.get('array_station_id'))
        # cls._store['pd_index_attr'] = pd.Index(cls.get('array_attr'))

        # cls._store['sf_index_2D'] = sf.IndexHierarchy.from_product(
        #         cls.get('array_datetime'),
        #         cls.get('array_station_id')
        #         )
        # cls._store['sf_index_3D'] = sf.IndexHierarchy.from_product(
        #         cls.get('array_datetime'),
        #         cls.get('array_station_id'),
        #         cls.get('array_attr')
        #         )

        # cls._store['tuple_index_2D'] = tuple(array2d_to_tuples(cls.get('sf_index_2D').values))
        # cls._store['tuple_index_3D'] = tuple(array2d_to_tuples(cls.get('sf_index_3D').values))



        cls._store['bsy2D_dict_df'] = BuoySingleYear2D.to_pd_dict()

        # cls._store['bsy2D_panel_obj'] = BuoySingleYear2D.to_pd_panel_obj()

        cls._store['bsy2D_panel_float'] = BuoySingleYear2D.to_pd_panel_float()

        cls._store['bsy2D_np'] = BuoySingleYear2D.to_np()

        cls._store['bsy2D_xarray_ds'] = BuoySingleYear2D.to_xarray_float()

        cls._store['bsy2D_pd_mi_2D'] = BuoySingleYear2D.to_pd()

        cls._store['bsy2D_sf_ih_2D'] = BuoySingleYear2D.to_sf()

        cls._store['bsy2D_pd_mi_1D_obj'] = BuoySingleYear1D.to_pd_obj()

        cls._store['bsy2D_pd_mi_1D_float'] = BuoySingleYear1D.to_pd_float()

        cls._store['bsy2D_sf_ih_1D'] = BuoySingleYear1D.to_sf()



    @classmethod
    def get(cls, key: str) -> tp.Any:
        return cls._store[key]


#-------------------------------------------------------------------------------

class OuterSingleMiddleSingleInnerSingle(_PerfTestPanel):

    @classmethod
    def dict_df(cls) -> None:
        data = SampleData.get('bsy2D_dict_df')
        post = data[46253].loc['2018-12-17T10:00:00', 'WVHT']
        assert post == 0.91

    @classmethod
    def panel_float(cls) -> None:
        data = SampleData.get('bsy2D_panel_float')
        post = data[46253, '2018-12-17T10:00:00', 'WVHT']
        assert post == 0.91

    @classmethod
    def np(cls) -> None:
        data, maps = SampleData.get('bsy2D_np')
        # using dictionary lookups;
        post = data[maps['station_id'][46253], maps['datetime'][np.datetime64('2018-12-17T10:00')], maps['attr']['WVHT']]
        assert post == 0.91

    @classmethod
    def xarray_da(cls) -> None:
        data = SampleData.get('bsy2D_xarray_ds')
        post = data.loc[46253, '2018-12-17T10:00:00', 'WVHT']
        assert float(post) == 0.91

    @classmethod
    def pd_mi_2D(cls) -> None:
        data = SampleData.get('bsy2D_pd_mi_2D')
        post = data.loc[pd.IndexSlice[46253, '2018-12-17T10:00:00'], 'WVHT']
        assert post.values[0] == 0.91

    @classmethod
    def sf_ih_2D(cls) -> None:
        data = SampleData.get('bsy2D_sf_ih_2D')
        post = data.loc[sf.HLoc[46253, '2018-12-17T10:00:00'], 'WVHT']
        assert post.values[0] == 0.91

    @classmethod
    def pd_mi_1D(cls) -> None:
        data = SampleData.get('bsy2D_pd_mi_1D_float')
        post = data.loc[46253, '2018-12-17T10:00:00', 'WVHT']
        assert post == 0.91

    @classmethod
    def sf_ih_1D(cls) -> None:
        data = SampleData.get('bsy2D_sf_ih_1D')
        post = data[sf.HLoc[46253, '2018-12-17T10:00:00', 'WVHT']]
        assert post == 0.91




class OuterSingleMiddleSliceInnerSingleMean(_PerfTestPanel):

    @classmethod
    def dict_df(cls) -> None:
        data = SampleData.get('bsy2D_dict_df')
        post = data[46253].loc['2018-12-17', 'WVHT'].mean()
        np.testing.assert_almost_equal(post, 1.2519148936170212)


    @classmethod
    def panel_float(cls) -> None:
        data = SampleData.get('bsy2D_panel_float')
        post = data[46253, '2018-12-17', 'WVHT'].mean()
        np.testing.assert_almost_equal(post, 1.2519148936170212)


    @classmethod
    def np(cls) -> None:
        # this is a float array
        data, maps = SampleData.get('bsy2D_np')
        # using dictionary lookups;
        post = np.nanmean(data[maps['station_id'][46253],
                maps['datetime'][np.datetime64('2018-12-17T00:00')]:maps['datetime'][np.datetime64('2018-12-18T00:00')],
                maps['attr']['WVHT']])
        np.testing.assert_almost_equal(post, 1.2519148936170212)



    @classmethod
    def xarray_da(cls) -> None:
        data = SampleData.get('bsy2D_xarray_ds')
        post = data.loc[46253, '2018-12-17', 'WVHT'].mean()
        np.testing.assert_almost_equal(float(post), 1.2519148936170212)



    @classmethod
    def pd_mi_2D(cls) -> None:
        data = SampleData.get('bsy2D_pd_mi_2D')
        post = data.loc[pd.IndexSlice[46253, '2018-12-17'], 'WVHT'].mean()
        np.testing.assert_almost_equal(post, 1.2519148936170212)

    @classmethod
    def sf_ih_2D(cls) -> None:
        data = SampleData.get('bsy2D_sf_ih_2D')
        post = data.loc[sf.HLoc[46253, '2018-12-17'], 'WVHT'].mean()
        np.testing.assert_almost_equal(post, 1.2519148936170212)



    @classmethod
    def pd_mi_1D(cls) -> None:
        data = SampleData.get('bsy2D_pd_mi_1D_float')
        # NOTE: could not get partial selection to work
        # pandas.core.indexing.IndexingError: Too many indexers
        post = data.loc[46253, '2018-12-17T00:00:00':'2018-12-17T23:30:00', 'WVHT'].mean()
        np.testing.assert_almost_equal(post, 1.2519148936170212)

    @classmethod
    def sf_ih_1D(cls) -> None:
        data = SampleData.get('bsy2D_sf_ih_1D')
        post = data[sf.HLoc[46253, '2018-12-17', 'WVHT']].mean()
        # post = data.loc[sf.HLoc[46253, '2018-12-17T00:00:00':'2018-12-17T23:30:00', 'WVHT']].mean()
        np.testing.assert_almost_equal(post, 1.2519148936170212)




class OuterAllMiddleSliceInnerSelectionMax(_PerfTestPanel):

    @classmethod
    def dict_df(cls) -> None:
        data = SampleData.get('bsy2D_dict_df')
        post = pd.concat([data[k].loc['2018-12-17'] for k in data.keys()])[['DPD', 'WVHT']].max()
        assert tuple(post.items()), (('DPD', 22.22), ('WVHT', 2.64))


    @classmethod
    def panel_float(cls) -> None:
        data = SampleData.get('bsy2D_panel_float')
        post = data[:, '2018-12-17'].max().max(axis=1)
        assert tuple(post.items()), (('DPD', 22.22), ('WVHT', 2.64))


    @classmethod
    def np(cls) -> None:
        # this is a float array
        data, maps = SampleData.get('bsy2D_np')
        date_slice = slice(maps['datetime'][np.datetime64('2018-12-17T00:00')],
                maps['datetime'][np.datetime64('2018-12-18T00:00')])
        post = np.vstack([data[maps['station_id'][k], date_slice] for k in maps['station_id'].keys()]).max(axis=0)
        assert post.tolist(), [22.22, 2.64]


    @classmethod
    def xarray_da(cls) -> None:
        data = SampleData.get('bsy2D_xarray_ds')
        post = data.loc[:, '2018-12-17'].max(axis=1).max(axis=0).values
        assert post.tolist(), [22.22, 2.64]



    @classmethod
    def pd_mi_2D(cls) -> None:
        data = SampleData.get('bsy2D_pd_mi_2D')
        post = data.loc[pd.IndexSlice[:, '2018-12-17'], ['WVHT', 'DPD']].max()
        assert tuple(post.items()), (('DPD', 22.22), ('WVHT', 2.64))

    @classmethod
    def sf_ih_2D(cls) -> None:
        data = SampleData.get('bsy2D_sf_ih_2D')
        post = data.loc[sf.HLoc[:, '2018-12-17'], ['WVHT', 'DPD']].max()
        assert tuple(post.items()), (('DPD', 22.22), ('WVHT', 2.64))



    @classmethod
    def pd_mi_1D(cls) -> None:
        data = SampleData.get('bsy2D_pd_mi_1D_float')
        post = tuple((attr, data.loc[:, '2018-12-17T00:00:00':'2018-12-17T23:30:00', attr].max()) for attr in ('DPD', 'WVHT'))
        assert post, (('DPD', 22.22), ('WVHT', 2.64))

    @classmethod
    def sf_ih_1D(cls) -> None:
        data = SampleData.get('bsy2D_sf_ih_1D')

        post = tuple((attr, data[sf.HLoc[:, '2018-12-17', attr]].max()) for attr in ('DPD', 'WVHT'))
        # post = data.loc[sf.HLoc[46253, '2018-12-17T00:00:00':'2018-12-17T23:30:00', 'WVHT']].mean()
        assert post, (('DPD', 22.22), ('WVHT', 2.64))



#-------------------------------------------------------------------------------
# import typing as tp


class BestDefense:

    @staticmethod
    def processor(
            functions: tp.Iterable[tp.Callable[[tp.Mapping[str, pd.Series]], pd.Series]],
            init: pd.Series
            ):

        results = {'init': init}

        for func in functions:
            results[func.__name__] = func(results.copy())

        return results

    @classmethod
    def processor_run(cls):

        def square(vm: tp.Mapping[str, pd.Series]):
            return vm['init'] ** 2

        def invert(vm: tp.Mapping[str, pd.Series]):
            return -vm['square']

        init = pd.Series((5, 2, 6), index=tuple('abc'))
        print(cls.processor(functions=[square, invert], init=init))


    class DataInterface:

        def __init__(self, size: int):
            self._square = pd.Series(np.arange(size) ** 2)
            self._invert = -self._square

        @property
        def square(self) -> pd.Series:
            return self._square.copy()

        @property
        def invert(self) -> pd.Series:
            return self._invert.copy()

    @classmethod
    def data_interface_run(cls):

        di = cls.DataInterface(6)
        print(di.square)
        print(di.invert)




#-------------------------------------------------------------------------------


if __name__ == '__main__':

    BestDefense.data_interface_run()


#     # fsf = BuoyLoader.buoy_to_sf(BUOYS[0], 2018)
#     fpd = BuoyLoader.buoy_to_pd(BUOYS[0], 2018)

#     #-----------------------------------------------------------
#     dfs = BuoySingleYear2D.to_pd_dict()
#     # ipdb> {station_id: df.loc['2018-12-17', 'WVHT'].mean() for station_id, df in dfs.items()}
#     # {46222: 1.5556249999999998, 46253: 1.2519148936170212, 46221: 1.6397916666666665}

#     # ipdb> pd.DataFrame.from_records(([station_id,] + df[['WVHT', 'DPD']].mean().tolist() for station_id, df in dfs.items()), columns=('station_id', 'WVHT', 'DPD'))
#     #    station_id      WVHT        DPD
#     # 0       46222  0.970099  11.813405
#     # 1       46253  0.927338  12.479156
#     # 2       46221  1.012615  12.883838


#     pna = BuoySingleYear2D.to_pd_panel_obj()
#     # <class 'pandas.core.panel.Panel'>
#     # Dimensions: 3 (items) x 17034 (major_axis) x 3 (minor_axis)
#     # Items axis: 46222 to 46221
#     # Major_axis axis: 2018-01-01 00:00:00 to 2018-12-31 23:30:00
#     # Minor_axis axis: DPD to MWD

#     # pna[:, '2018-12-17', ['WVHT', 'DPD']].mean()
#     #           46222      46253      46221
#     # WVHT   1.555625   1.251915   1.639792
#     # DPD   14.927500  14.190851  14.811667

#     # ipdb> pna[46222, :, 'DPD'].values
#     # array([11.76, 10.53, 11.11, ..., 15.38, 15.38, 15.38], dtype=object)

#     # ipdb> pna[46222].dtypes
#     # DPD     object
#     # WVHT    object
#     # MWD     object
#     # dtype: object

#     # ipdb> pna.values.shape
#     # (3, 17034, 3)

#     pnb = BuoySingleYear2D.to_pd_panel_float()

#     # ipdb> pnb
#     # <class 'pandas.core.panel.Panel'>
#     # Dimensions: 3 (items) x 17034 (major_axis) x 2 (minor_axis)
#     # Items axis: 46222 to 46221
#     # Major_axis axis: 2018-01-01 00:00:00 to 2018-12-31 23:30:00
#     # Minor_axis axis: DPD to WVHT

#     fpd = BuoySingleYear2D.to_pd()

#     # ipdb> fpd.loc[pd.IndexSlice[46221, datetime.datetime(2018, 12, 16, 10, 30)], 'WVHT']
#     #1.47

# # ipdb> fpd.loc[pd.IndexSlice[46221, '2018-12-17'], 'WVHT']
#     # *** pandas.errors.UnsortedIndexError: 'MultiIndex slicing requires the index to be lexsorted: slicing on levels [1], lexsort depth 0'

#     # >>> fpd.sort_index(inplace=True)

#     # ipdb> fpd.loc[pd.IndexSlice[46221, '2018-12-17'], 'WVHT'].head()
#     # station_id  datetime
#     # 46221       2018-12-17 00:00:00    1.43
#     #             2018-12-17 00:30:00    1.32
#     #             2018-12-17 01:00:00    1.38
#     #             2018-12-17 01:30:00    1.42
#     #             2018-12-17 02:00:00    1.36
#     # Name: WVHT, dtype: float64

#     # ipdb> fpd.loc[pd.IndexSlice[:, '2018-12-17'], ['WVHT', 'DPD']].mean()
#     # WVHT     1.484056
#     # DPD     14.646503
#     # dtype: float64
#     # ipdb> fpd.loc[pd.IndexSlice[:, '2018-12-18'], ['WVHT', 'DPD']].mean()
#     # WVHT     2.155594
#     # DPD     16.404965
#     # dtype: float64



#     fsf = BuoySingleYear2D.to_sf()

#     BuoySingleYear2D.process_sf()



#     # BuoySingleYear2D.process_np()
#     # BuoySingleYear2D.process_pd_panel()


#     # ssf = BuoySingleYear1D.to_sf()
#     spda = BuoySingleYear1D.to_pd_obj()
#     spdb = BuoySingleYear1D.to_pd_float()

#     # df = BuoySingleYear2D.process_pd_multi_index()













# example of pandas series that supports partial matching
# In : s1 = pd.Series(range(60), index=pd.date_range('1999-12', freq='D', periods=60))

# s1 = pd.Series(range(120), pd.MultiIndex.from_product((('a', 'b'), pd.date_range('1999-12', freq='D', periods=60))))
# # this does not work
# s1[pd.IndexSlice['a', '2000']]


# partial date string matching
# https://github.com/pandas-dev/pandas/issues/25165


# class BuoyMultiYear:

#     @staticmethod
#     def to_sf(years: tp.Iterable[int] = (2017, 2017, 2018)) -> sf.Frame:

#         frames = []
#         for buoy in BUOYS:
#             for year in years:
#                 f = BuoyLoader.buoy_to_sf(buoy, year).to_frame_go()
#                 f['year'] = year
#                 f['month'] = [int(x.split('-')[1]) for x in f['datetime'].values]
#                 # import ipdb; ipdb.set_trace()
#                 frames.append(f)

#         f = sf.Frame.from_concat(frames,
#                 axis=0,
#                 index=sf.IndexAutoFactory,
#                 name='buos_multi_year'
#                 )
#         # NOTE: this fails unexpectedly
#         f = f.set_index_hierarchy(('station_id', 'datetime'),
#                 index_constructors=(sf.Index, sf.IndexMinute),
#                 drop=True)
#         return f





import unittest
import typing as tp

import numpy as np

from static_frame.core.frame import Frame
from static_frame.core.hloc import HLoc
from static_frame.core.index_hierarchy import IndexHierarchy
from static_frame.core.store import StoreConfig
from static_frame.core.store import StoreConfigMap
from static_frame.core.store_filter import StoreFilter
from static_frame.core.store_xlsx import StoreXLSX
from static_frame.core.util import STORE_LABEL_DEFAULT
from static_frame.test.test_case import temp_file
from static_frame.test.test_case import TestCase

class TestUnit(TestCase):


    def test_store_xlsx_write_a(self) -> None:

        f1 = Frame.from_dict(
                dict(x=(1,2,-5,200), y=(3,4,-5,-3000)),
                index=IndexHierarchy.from_product(('I', 'II'), ('a', 'b')),
                name='f1')
        f2 = Frame.from_dict(
                dict(a=(1,2,3), b=(4,5,6)),
                index=('x', 'y', 'z'),
                name='f2')
        f3 = Frame.from_records(
                ((10, 20, 50, 60), (50.0, 60.4, -50, -60)),
                index=('p', 'q'),
                columns=IndexHierarchy.from_product(('I', 'II'), ('a', 'b')),
                name='f3')
        f4 = Frame.from_records((
                (10, 20, 50, False, 10, 20, 50, False),
                (50.0, 60.4, -50, True, 50.0, 60.4, -50, True),
                (234, 44452, 0, False, 234, 44452, 0, False),
                (4, -4, 2000, True, 4, -4, 2000, True),
                (10, 20, 50, False, 10, 20, 50, False),
                (50.0, 60.4, -50, True, 50.0, 60.4, -50, True),
                (234, 44452, 0, False, 234, 44452, 0, False),
                (4, -4, 2000, True, 4, -4, 2000, True),
                ),
                index=IndexHierarchy.from_product(('top', 'bottom'), ('far', 'near'), ('left', 'right')),
                columns=IndexHierarchy.from_product(('I', 'II'), ('a', 'b'), (1, 2)),
                name='f4')

        frames = (f1, f2, f3, f4)
        config_map = StoreConfigMap.from_config(
                StoreConfig(include_index=True, include_columns=True))

        with temp_file('.xlsx') as fp:

            st1 = StoreXLSX(fp)
            st1.write(((f.name, f) for f in frames), config=config_map)

            sheet_names = tuple(st1.labels()) # this will read from file, not in memory
            self.assertEqual(tuple(f.name for f in frames), sheet_names)

            for i, name in enumerate(sheet_names):
                f_src = frames[i]
                c = StoreConfig(
                        index_depth=f_src.index.depth,
                        columns_depth=f_src.columns.depth
                        )
                f_loaded = st1.read(name, config=c)
                self.assertEqualFrames(f_src, f_loaded, compare_dtype=False)


    def test_store_xlsx_write_b(self) -> None:

        f1 = Frame.from_records(
                ((None, np.nan, 50, 'a'), (None, -np.inf, -50, 'b'), (None, 60.4, -50, 'c')),
                index=('p', 'q', 'r'),
                columns=IndexHierarchy.from_product(('I', 'II'), ('a', 'b')),
                )

        config_map = StoreConfigMap.from_config(
                StoreConfig(include_index=True, include_columns=True))

        with temp_file('.xlsx') as fp:

            st = StoreXLSX(fp)
            st.write(((STORE_LABEL_DEFAULT, f1),), config=config_map)

            c = StoreConfig(
                    index_depth=f1.index.depth,
                    columns_depth=f1.columns.depth
                    )
            f2 = st.read(STORE_LABEL_DEFAULT, config=c)

            # just a sample column for now
            self.assertEqual(
                    f1[HLoc[('II', 'a')]].values.tolist(),
                    f2[HLoc[('II', 'a')]].values.tolist() )

            self.assertEqualFrames(f1, f2)


    def test_store_xlsx_read_a(self) -> None:
        f1 = Frame.from_elements([1, 2, 3], index=('a', 'b', 'c'), columns=('x',))

        config_map = StoreConfigMap.from_config(
                StoreConfig(include_index=False, include_columns=True))

        with temp_file('.xlsx') as fp:

            st = StoreXLSX(fp)
            st.write(((STORE_LABEL_DEFAULT, f1),), config=config_map)

            c = StoreConfig(
                    index_depth=0,
                    columns_depth=f1.columns.depth
                    )
            f2 = st.read(STORE_LABEL_DEFAULT, config=c)

        self.assertTrue((f1.values == f2.values).all())
        self.assertEqual(f2.to_pairs(0),
                (('x', ((0, 1), (1, 2), (2, 3))),)
                )


    def test_store_xlsx_read_b(self) -> None:
        index = IndexHierarchy.from_product(('left', 'right'), ('up', 'down'))
        columns = IndexHierarchy.from_labels(((100, -5, 20),))

        f1 = Frame.from_elements([1, 2, 3, 4], index=index, columns=columns)

        config_map = StoreConfigMap.from_config(
                StoreConfig(include_index=False, include_columns=True))

        with temp_file('.xlsx') as fp:

            st = StoreXLSX(fp)
            st.write(((STORE_LABEL_DEFAULT, f1),), config=config_map)

            c = StoreConfig(
                    index_depth=0,
                    columns_depth=f1.columns.depth
                    )
            f2 = st.read(STORE_LABEL_DEFAULT, config=c)

        self.assertTrue((f1.values == f2.values).all())
        self.assertEqual(f2.to_pairs(0),
                (((100, -5, 20), ((0, 1), (1, 2), (2, 3), (3, 4))),)
                )


    def test_store_xlsx_read_c(self) -> None:
        index = IndexHierarchy.from_product(('left', 'right'), ('up', 'down'))
        columns = IndexHierarchy.from_labels(((100, -5, 20),))

        f1 = Frame.from_elements([1, 2, 3, 4], index=index, columns=columns)

        config_map = StoreConfigMap.from_config(
                StoreConfig(include_index=True, include_columns=False))

        with temp_file('.xlsx') as fp:

            st = StoreXLSX(fp)
            st.write(((STORE_LABEL_DEFAULT, f1),), config=config_map[None])

            c = StoreConfig(
                    index_depth=f1.index.depth,
                    columns_depth=0
                    )
            f2 = st.read(STORE_LABEL_DEFAULT, config=c)

        self.assertTrue((f1.values == f2.values).all())
        self.assertEqual(f2.to_pairs(0),
                ((0, ((('left', 'up'), 1), (('left', 'down'), 2), (('right', 'up'), 3), (('right', 'down'), 4))),)
                )


    def test_store_xlsx_read_d(self) -> None:

        f1 = Frame.from_records(
                ((10, 20, 50, 60), (50.0, 60.4, -50, -60)),
                index=('p', 'q'),
                columns=('a', 'b', 'c', 'd'),
                name='f1')

        sc1 = StoreConfig(include_index=False, include_columns=True)
        sc2 = StoreConfig(columns_depth=0, index_depth=0)

        with temp_file('.xlsx') as fp:

            st = StoreXLSX(fp)
            st.write(((STORE_LABEL_DEFAULT, f1),), config=sc1)


            f2 = st.read(STORE_LABEL_DEFAULT) #  get default config
            self.assertEqual(f2.to_pairs(0),
                    (('a', ((0, 10), (1, 50))), ('b', ((0, 20.0), (1, 60.4))), ('c', ((0, 50), (1, -50))), ('d', ((0, 60), (1, -60)))))

            f3 = st.read(STORE_LABEL_DEFAULT, config=sc2)
            self.assertEqual(f3.to_pairs(0),
                    ((0, ((0, 'a'), (1, 10), (2, 50))), (1, ((0, 'b'), (1, 20), (2, 60.4))), (2, ((0, 'c'), (1, 50), (2, -50))), (3, ((0, 'd'), (1, 60), (2, -60)))))


    def test_store_xlsx_read_e(self) -> None:

        f1 = Frame.from_records(
                ((np.inf, np.inf), (-np.inf, -np.inf)),
                index=('p', 'q'),
                columns=('a', 'b'),
                name='f1')

        sc1 = StoreConfig(columns_depth=1, index_depth=1)

        with temp_file('.xlsx') as fp:

            st = StoreXLSX(fp)
            st.write(((STORE_LABEL_DEFAULT, f1),))

            f1 = st.read(STORE_LABEL_DEFAULT, config=sc1, store_filter=None)
            self.assertEqual(f1.to_pairs(0),
                    (('a', (('p', 'inf'), ('q', '-inf'))), ('b', (('p', 'inf'), ('q', '-inf')))))

            f2 = st.read(STORE_LABEL_DEFAULT, config=sc1, store_filter=StoreFilter())
            self.assertEqual(f2.to_pairs(0),
                    (('a', (('p', np.inf), ('q', -np.inf))),
                    ('b', (('p', np.inf), ('q', -np.inf)))))

    def test_store_xlsx_read_many_a(self) -> None:

        f1 = Frame.from_dict(
                dict(x=(1,2,-5,200), y=(3,4,-5,-3000)),
                index=IndexHierarchy.from_product(('I', 'II'), ('a', 'b')),
                name='f1')
        f2 = Frame.from_dict(
                dict(a=(1,2,3), b=(4,5,6)),
                index=('x', 'y', 'z'),
                name='f2')
        f3 = Frame.from_records(
                ((10, 20, 50, 60), (50.0, 60.4, -50, -60)),
                index=('p', 'q'),
                columns=IndexHierarchy.from_product(('I', 'II'), ('a', 'b')),
                name='f3')
        f4 = Frame.from_records((
                (10, 20, 50, False, 10, 20, 50, False),
                (50.0, 60.4, -50, True, 50.0, 60.4, -50, True),
                (234, 44452, 0, False, 234, 44452, 0, False),
                (4, -4, 2000, True, 4, -4, 2000, True),
                (10, 20, 50, False, 10, 20, 50, False),
                (50.0, 60.4, -50, True, 50.0, 60.4, -50, True),
                (234, 44452, 0, False, 234, 44452, 0, False),
                (4, -4, 2000, True, 4, -4, 2000, True),
                ),
                index=IndexHierarchy.from_product(('top', 'bottom'), ('far', 'near'), ('left', 'right')),
                columns=IndexHierarchy.from_product(('I', 'II'), ('a', 'b'), (1, 2)),
                name='f4')

        frames = (f1, f2, f3, f4)
        config_map_write = StoreConfigMap.from_config(
                StoreConfig(include_index=True, include_columns=True))

        with temp_file('.xlsx') as fp:

            st1 = StoreXLSX(fp)
            st1.write(((f.name, f) for f in frames), config=config_map_write)

            sheet_names = tuple(st1.labels()) # this will read from file, not in memory
            self.assertEqual(tuple(f.name for f in frames), sheet_names)

            config_map_read: tp.Dict[tp.Hashable, StoreConfig] = {}
            for i, name in enumerate(sheet_names):
                f_src = frames[i]
                c = StoreConfig(
                        index_depth=f_src.index.depth,
                        columns_depth=f_src.columns.depth
                        )
                config_map_read[name] = c

            for i, f_loaded in enumerate(st1.read_many(sheet_names, config=config_map_read)):
                f_src = frames[i]
                self.assertEqualFrames(f_src, f_loaded, compare_dtype=False)



    #---------------------------------------------------------------------------

    def test_dtype_to_writer_attr(self) -> None:
        attr1, switch1 = StoreXLSX._dtype_to_writer_attr(
                np.array(('2020', '2021'), dtype=np.datetime64).dtype)

        self.assertEqual(attr1, 'write')
        self.assertEqual(switch1, True)

        attr2, switch2 = StoreXLSX._dtype_to_writer_attr(
                np.array(('2020-01-01', '2021-01-01'), dtype=np.datetime64).dtype)

        self.assertEqual(attr2, 'write')
        self.assertEqual(switch2, True)

if __name__ == '__main__':
    unittest.main()






import unittest

from static_frame.core.container import ContainerOperand
from static_frame.test.test_case import TestCase
from static_frame.core.interface import UFUNC_AXIS_SKIPNA
from static_frame.core.interface import UFUNC_SHAPE_SKIPNA


class TestUnit(TestCase):

    def test_container_attrs(self) -> None:

        for attr in UFUNC_AXIS_SKIPNA.keys() | UFUNC_SHAPE_SKIPNA.keys():
            c = ContainerOperand
            self.assertTrue(hasattr(c, attr))

        with self.assertRaises(NotImplementedError):
            c().display()




if __name__ == '__main__':
    unittest.main()




import unittest
import numpy as np
# import typing as tp

from static_frame.core.index_base import IndexBase

# from static_frame.core.exception import ErrorInitIndex
from static_frame.test.test_case import TestCase



class TestUnit(TestCase):

    def test_index_base_slotted_a(self) -> None:
        idx1 = IndexBase()

        with self.assertRaises(AttributeError):
            idx1.g = 30 # type: ignore #pylint: disable=E0237
        with self.assertRaises(AttributeError):
            idx1.__dict__ #pylint: disable=W0104

    def test_index_base_not_implemented(self) -> None:

        idx1 = IndexBase()

        with self.assertRaises(NotImplementedError):
            idx1._ufunc_axis_skipna(axis=0,
                    skipna=False,
                    ufunc=np.sum,
                    ufunc_skipna=np.nansum,
                    composable=True,
                    dtypes=(),
                    size_one_unity=True)


        with self.assertRaises(NotImplementedError):
            idx1._update_array_cache()

        with self.assertRaises(NotImplementedError):
            idx1.copy()

        with self.assertRaises(NotImplementedError):
            idx1.copy()

        with self.assertRaises(NotImplementedError):
            idx1.display()

        with self.assertRaises(NotImplementedError):
            idx1.from_labels(())



if __name__ == '__main__':
    unittest.main()





import typing as tp
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor

import numpy as np

from static_frame.core.bus import Bus
from static_frame.core.container import ContainerOperand
from static_frame.core.display import Display
from static_frame.core.display import DisplayActive
from static_frame.core.display import DisplayHeader
from static_frame.core.display_config import DisplayConfig
from static_frame.core.doc_str import doc_inject
from static_frame.core.frame import Frame
from static_frame.core.index_auto import IndexAutoFactoryType
from static_frame.core.node_selector import InterfaceGetItem
from static_frame.core.node_selector import InterfaceSelectTrio
from static_frame.core.series import Series
from static_frame.core.store import Store
from static_frame.core.store import StoreConfigMap
from static_frame.core.store import StoreConfigMapInitializer
from static_frame.core.store_client_mixin import StoreClientMixin
from static_frame.core.store_hdf5 import StoreHDF5
from static_frame.core.store_sqlite import StoreSQLite
from static_frame.core.store_xlsx import StoreXLSX
from static_frame.core.store_zip import StoreZipCSV
from static_frame.core.store_zip import StoreZipParquet
from static_frame.core.store_zip import StoreZipPickle
from static_frame.core.store_zip import StoreZipTSV
from static_frame.core.util import AnyCallable
from static_frame.core.util import Bloc2DKeyType
from static_frame.core.util import DEFAULT_SORT_KIND
from static_frame.core.util import DTYPE_OBJECT
from static_frame.core.util import ELEMENT_TUPLE
from static_frame.core.util import GetItemKeyType
from static_frame.core.util import GetItemKeyTypeCompound
from static_frame.core.util import IndexInitializer
from static_frame.core.util import KeyOrKeys
from static_frame.core.util import NameType
from static_frame.core.util import PathSpecifier
from static_frame.core.util import UFunc
from static_frame.core.style_config import StyleConfig


FrameOrSeries = tp.Union[Frame, Series]
IteratorFrameItems = tp.Iterator[tp.Tuple[tp.Hashable, FrameOrSeries]]
GeneratorFrameItems = tp.Callable[..., IteratorFrameItems]


#-------------------------------------------------------------------------------
# family of executor functions normalized in signature (taking a single tuple of args) for usage in processor pool calls

def normalize_container(post: tp.Any
        ) -> FrameOrSeries:
    # post might be an element, promote to a Series to permit concatenation
    if post.__class__ is np.ndarray:
        if post.ndim == 1:
            return Series(post)
        elif post.ndim == 2:
            return Frame(post)
        # let ndim 0 pass
    if not isinstance(post, (Frame, Series)):
        # NOTE: do not set index as (container.name,), as this can lead to diagonal formations; will already be paired with stored labels
        return Series.from_element(post, index=ELEMENT_TUPLE)
    return post

def call_func(bundle: tp.Tuple[FrameOrSeries, AnyCallable]
        ) -> FrameOrSeries:
    container, func = bundle
    return normalize_container(func(container))

def call_func_items(bundle: tp.Tuple[FrameOrSeries, AnyCallable, tp.Hashable]
        ) -> FrameOrSeries:
    container, func, label = bundle
    return normalize_container(func(label, container))

def call_attr(bundle: tp.Tuple[FrameOrSeries, str, tp.Any, tp.Any]
        ) -> FrameOrSeries:
    container, attr, args, kwargs = bundle
    func = getattr(container, attr)
    return normalize_container(func(*args, **kwargs))

#-------------------------------------------------------------------------------
class Batch(ContainerOperand, StoreClientMixin):
    '''
    A lazy, sequentially evaluated container of :obj:`Frame` that broadcasts operations on contained :obj:`Frame` by return new :obj:`Batch` instances. Full evaluation of operations only occurs when iterating or calling an exporter.
    '''

    __slots__ = (
            '_items',
            '_name',
            '_config',
            '_max_workers',
            '_chunksize',
            '_use_threads',
            )

    _config: StoreConfigMap

    @classmethod
    def from_frames(cls,
            frames: tp.Iterable[Frame],
            *,
            name: NameType = None,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''Return a :obj:`Batch` from an iterable of :obj:`Frame`; labels will be drawn from :obj:`Frame.name`.
        '''
        return cls(((f.name, f) for f in frames),
                name=name,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                )

    #---------------------------------------------------------------------------
    # constructors by data format

    @classmethod
    def _from_store(cls,
            store: Store,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        config_map = StoreConfigMap.from_initializer(config)

        items = ((label, store.read(label, config=config_map[label]))
                for label in store.labels(config=config_map))

        return cls(items,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )


    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_zip_tsv(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to zipped TSV :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreZipTSV(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )

    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_zip_csv(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to zipped CSV :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreZipCSV(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )

    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_zip_pickle(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to zipped pickle :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreZipPickle(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )


    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_zip_parquet(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to zipped parquet :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreZipParquet(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )


    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_xlsx(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to an XLSX :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        # how to pass configuration for multiple sheets?
        store = StoreXLSX(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )


    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_sqlite(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to an SQLite :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreSQLite(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )


    @classmethod
    @doc_inject(selector='batch_constructor')
    def from_hdf5(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> 'Batch':
        '''
        Given a file path to a HDF5 :obj:`Batch` store, return a :obj:`Batch` instance.

        {args}
        '''
        store = StoreHDF5(fp)
        return cls._from_store(store,
                config=config,
                max_workers=max_workers,
                chunksize=chunksize,
                use_threads=use_threads,
                                )

    #---------------------------------------------------------------------------

    @doc_inject(selector='batch_init')
    def __init__(self,
            items: IteratorFrameItems,
            *,
            name: NameType = None,
            config: StoreConfigMapInitializer = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ):
        '''
        Default constructor of a :obj:`Batch`.

        {args}
        '''
        self._items = items # might be a generator!
        self._name = name

        self._config = StoreConfigMap.from_initializer(config)

        self._max_workers = max_workers
        self._chunksize = chunksize
        self._use_threads = use_threads

    #---------------------------------------------------------------------------
    def _derive(self,
            gen: GeneratorFrameItems,
            name: NameType = None,
            ) -> 'Batch':
        '''Utility for creating derived Batch
        '''
        return self.__class__(gen(),
                name=name if name is not None else self._name,
                config=self._config,
                max_workers=self._max_workers,
                chunksize=self._chunksize,
                use_threads=self._use_threads,
                )

    #---------------------------------------------------------------------------
    # name interface

    @property #type: ignore
    @doc_inject()
    def name(self) -> NameType:
        '''{}'''
        return self._name

    def rename(self, name: NameType) -> 'Batch':
        '''
        Return a new Batch with an updated name attribute.
        '''
        def gen() -> IteratorFrameItems:
            yield from self._items
        return self._derive(gen, name=name)

    #---------------------------------------------------------------------------
    @property
    def shapes(self) -> Series:
        '''A :obj:`Series` describing the shape of each iterated :obj:`Frame`.

        Returns:
            :obj:`tp.Tuple[int]`
        '''
        items = ((label, f.shape) for label, f in self._items)
        return Series.from_items(items, name='shape', dtype=DTYPE_OBJECT)


    def display(self,
            config: tp.Optional[DisplayConfig] = None,
            *,
            style_config: tp.Optional[StyleConfig] = None,
            ) -> Display:
        '''Provide a :obj:`Series`-style display of the :obj:`Batch`. Note that if the held iterator is a generator, this display will exhaust the generator.
        '''
        config = config or DisplayActive.get()

        items = ((label, f.__class__) for label, f in self._items)
        series = Series.from_items(items, name=self._name)

        display_cls = Display.from_values((),
                header=DisplayHeader(self.__class__, self._name),
                config=config)
        return series._display(config,
                display_cls=display_cls,
                style_config=style_config,
                )

    def __repr__(self) -> str:
        '''Provide a display of the :obj:`Batch` that does not exhaust the generator.
        '''
        if self._name:
            header = f'{self.__class__.__name__}: {self._name}'
        else:
            header = self.__class__.__name__
        return f'<{header} at {hex(id(self))}>'

    #---------------------------------------------------------------------------
    # core function application routines

    def _apply_pool(self,
            labels: tp.List[tp.Hashable],
            arg_iter: tp.Iterator[tp.Tuple[tp.Any, ...]],
            caller: tp.Callable[..., FrameOrSeries],
            ) -> 'Batch':

        pool_executor = ThreadPoolExecutor if self._use_threads else ProcessPoolExecutor

        def gen_pool() -> IteratorFrameItems:
            with pool_executor(max_workers=self._max_workers) as executor:
                yield from zip(labels,
                        executor.map(caller, arg_iter, chunksize=self._chunksize)
                        )
        return self._derive(gen_pool)

    def _apply_pool_except(self,
            labels: tp.List[tp.Hashable],
            arg_iter: tp.Iterator[tp.Tuple[tp.Any, ...]],
            caller: tp.Callable[..., FrameOrSeries],
            exception: tp.Type[Exception],
            ) -> 'Batch':

        if self._chunksize != 1:
            raise NotImplementedError('Cannot use apply_except idioms with chunksize other than 1')

        pool_executor = ThreadPoolExecutor if self._use_threads else ProcessPoolExecutor

        def gen_pool() -> IteratorFrameItems:
            futures = []
            with pool_executor(max_workers=self._max_workers) as executor:
                for args in arg_iter:
                    futures.append(executor.submit(caller, args))

                for label, future in zip(labels, futures):
                    try:
                        container = future.result()
                    except exception:
                        continue
                    yield label, container

        return self._derive(gen_pool)

    def _apply_attr(self,
            *args: tp.Any,
            attr: str,
            **kwargs: tp.Any,
            ) -> 'Batch':
        '''
        Apply a method on a Frame given as an attr string.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._items:
                    yield label, call_attr((frame, attr, args, kwargs))
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, str, tp.Any, tp.Any]]:
            for label, frame in self._items:
                labels.append(label)
                yield frame, attr, args, kwargs

        return self._apply_pool(labels, arg_gen(), call_attr)

    def apply(self, func: AnyCallable) -> 'Batch':
        '''
        Apply a function to each :obj:`Frame` contained in this :obj:`Frame`, where a function is given the :obj:`Frame` as an argument.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._items:
                    yield label, call_func((frame, func))
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, AnyCallable]]:
            for label, frame in self._items:
                labels.append(label)
                yield frame, func

        return self._apply_pool(labels, arg_gen(), call_func)

    def apply_except(self,
            func: AnyCallable,
            exception: tp.Type[Exception],
            ) -> 'Batch':
        '''
        Apply a function to each :obj:`Frame` contained in this :obj:`Frame`, where a function is given the :obj:`Frame` as an argument. Exceptions raised that matching the `except` argument will be silenced.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._items:
                    try:
                        yield label, call_func((frame, func))
                    except exception:
                        pass
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, AnyCallable]]:
            for label, frame in self._items:
                labels.append(label)
                yield frame, func

        return self._apply_pool_except(labels,
                arg_gen(),
                call_func,
                exception,
                )

    def apply_items(self, func: AnyCallable) -> 'Batch':
        '''
        Apply a function to each :obj:`Frame` contained in this :obj:`Frame`, where a function is given the pair of label, :obj:`Frame` as an argument.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._items:
                    yield label, call_func_items((frame, func, label))
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, AnyCallable, tp.Hashable]]:
            for label, frame in self._items:
                labels.append(label)
                yield frame, func, label

        return self._apply_pool(labels, arg_gen(), call_func_items)

    def apply_items_except(self,
            func: AnyCallable,
            exception: tp.Type[Exception],
            ) -> 'Batch':
        '''
        Apply a function to each :obj:`Frame` contained in this :obj:`Frame`, where a function is given the pair of label, :obj:`Frame` as an argument. Exceptions raised that matching the `except` argument will be silenced.
        '''
        if self._max_workers is None:
            def gen() -> IteratorFrameItems:
                for label, frame in self._items:
                    try:
                        yield label, call_func_items((frame, func, label))
                    except exception:
                        pass
            return self._derive(gen)

        labels = []
        def arg_gen() -> tp.Iterator[tp.Tuple[FrameOrSeries, AnyCallable, tp.Hashable]]:
            for label, frame in self._items:
                labels.append(label)
                yield frame, func, label

        return self._apply_pool_except(labels,
                arg_gen(),
                call_func_items,
                exception,
                )

    #---------------------------------------------------------------------------
    # extraction

    def _extract_iloc(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_extract_iloc',
                key=key
                )
    def _extract_loc(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_extract_loc',
                key=key
                )

    def _extract_bloc(self, key: Bloc2DKeyType) -> 'Batch':
        return self._apply_attr(
                attr='_extract_bloc',
                key=key
                )

    def __getitem__(self, key: GetItemKeyType) -> 'Batch':
        ''
        return self._apply_attr(
                attr='__getitem__',
                key=key
                )

    #---------------------------------------------------------------------------
    def _drop_iloc(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_drop_iloc',
                key=key
                )

    def _drop_loc(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_drop_loc',
                key=key
                )

    def _drop_getitem(self, key: GetItemKeyTypeCompound) -> 'Batch':
        return self._apply_attr(
                attr='_drop_getitem',
                key=key
                )

    #---------------------------------------------------------------------------
    # interfaces

    @property
    def loc(self) -> InterfaceGetItem['Batch']:
        return InterfaceGetItem(self._extract_loc)

    @property
    def iloc(self) -> InterfaceGetItem['Batch']:
        return InterfaceGetItem(self._extract_iloc)

    @property
    def bloc(self) -> InterfaceGetItem['Batch']:
        return InterfaceGetItem(self._extract_bloc)

    @property
    def drop(self) -> InterfaceSelectTrio['Batch']:
        return InterfaceSelectTrio( #type: ignore
            func_iloc=self._drop_iloc,
            func_loc=self._drop_loc,
            func_getitem=self._drop_getitem)

    # NOTE: note sure if assign interfaces would work in this context

    #---------------------------------------------------------------------------
    # dictionary-like interface
    # these methods operate on the Batch itself, not the contained Frames

    def keys(self) -> tp.Iterator[tp.Hashable]:
        '''
        Iterator of :obj:`Frame` labels.
        '''
        for k, _ in self._items:
            yield k

    def __iter__(self) -> tp.Iterator[tp.Hashable]:
        '''
        Iterator of :obj:`Frame` labels, same as :obj:`Batch.keys`.
        '''
        yield from self.keys()

    @property
    def values(self) -> tp.Iterator[FrameOrSeries]:
        '''
        Return an iterator of values (:obj:`Frame` or :obj:`Series`) stored in this :obj:`Batch`.
        '''
        return (v for _, v in self._items)

    def items(self) -> IteratorFrameItems:
        '''
        Iterator of labels, :obj:`Frame`.
        '''
        return self._items.__iter__()

    _items_store = items

    #---------------------------------------------------------------------------
    # axis and shape ufunc methods

    def _ufunc_unary_operator(self,
            operator: UFunc
            ) -> 'Batch':
        return self._apply_attr(
                attr='_ufunc_unary_operator',
                operator=operator
                )

    def _ufunc_binary_operator(self, *,
            operator: UFunc,
            other: tp.Any,
            fill_value: object = np.nan,
            ) -> 'Batch':
        return self._apply_attr(
                attr='_ufunc_binary_operator',
                operator=operator,
                other=other,
                )

    def _ufunc_axis_skipna(self, *,
            axis: int,
            skipna: bool,
            ufunc: UFunc,
            ufunc_skipna: UFunc,
            composable: bool,
            dtypes: tp.Tuple[np.dtype, ...],
            size_one_unity: bool
            ) -> 'Batch':
        return self._apply_attr(
                attr='_ufunc_axis_skipna',
                axis=axis,
                skipna=skipna,
                ufunc=ufunc,
                ufunc_skipna=ufunc_skipna,
                composable=composable,
                dtypes=dtypes,
                size_one_unity=size_one_unity,
                )

    def _ufunc_shape_skipna(self, *,
            axis: int,
            skipna: bool,
            ufunc: UFunc,
            ufunc_skipna: UFunc,
            composable: bool,
            dtypes: tp.Tuple[np.dtype, ...],
            size_one_unity: bool
            ) -> 'Batch':

        return self._apply_attr(
                attr='_ufunc_shape_skipna',
                axis=axis,
                skipna=skipna,
                ufunc=ufunc,
                ufunc_skipna=ufunc_skipna,
                composable=composable,
                dtypes=dtypes,
                size_one_unity=size_one_unity,
                )

    #---------------------------------------------------------------------------
    # transformations resulting in the same dimensionality

    def sort_index(self,
            *,
            ascending: bool = True,
            kind: str = DEFAULT_SORT_KIND
            ) -> 'Batch':
        '''
        Return a new :obj:`Batch` with contained :obj;`Frame` ordered by the sorted ``index``.
        '''
        return self._apply_attr(
                attr='sort_index',
                ascending=ascending,
                kind=kind,
                )

    def sort_columns(self,
            *,
            ascending: bool = True,
            kind: str = DEFAULT_SORT_KIND
            ) -> 'Batch':
        '''
        Return a new :obj:`Batch` with contained :obj:`Frame` ordered by the sorted ``columns``.
        '''
        return self._apply_attr(
                attr='sort_columns',
                ascending=ascending,
                kind=kind,
                )

    def sort_values(self,
            label: KeyOrKeys,
            *,
            ascending: bool = True,
            axis: int = 1,
            kind: str = DEFAULT_SORT_KIND) -> 'Batch':
        '''
        Return a new :obj:`Batch` with contained :obj:`Frame` ordered by the sorted values, where values are given by single column or iterable of columns.

        Args:
            label: a label or iterable of keys.
        '''
        return self._apply_attr(
                attr='sort_values',
                label=label,
                ascending=ascending,
                axis=axis,
                kind=kind,
                )

    def isin(self, other: tp.Any) -> 'Batch':
        '''
        Return a new :obj:`Batch` with contained :obj:`Frame` as a same-sized Boolean :obj:`Frame` that shows if the same-positioned element is in the passed iterable.
        '''
        return self._apply_attr(
                attr='isin',
                other=other,
                )

    @doc_inject(class_name='Batch')
    def clip(self, *,
            lower: tp.Optional[tp.Union[float, Series, Frame]] = None,
            upper: tp.Optional[tp.Union[float, Series, Frame]] = None,
            axis: tp.Optional[int] = None
            ) -> 'Batch':
        '''{}

        Args:
            lower: value, :obj:`Series`, :obj:`Frame`
            upper: value, :obj:`Series`, :obj:`Frame`
            axis: required if ``lower`` or ``upper`` are given as a :obj:`Series`.
        '''
        return self._apply_attr(
                attr='clip',
                lower=lower,
                upper=upper,
                axis=axis,
                )

    def transpose(self) -> 'Batch':
        '''Transpose. Return a :obj:`Frame` with ``index`` as ``columns`` and vice versa.
        '''
        return self._apply_attr(attr='transpose')

    @property
    def T(self) -> 'Batch':
        '''Transpose. Return a :obj:`Frame` with ``index`` as ``columns`` and vice versa.
        '''
        return self._apply_attr(attr='transpose')



    @doc_inject(selector='duplicated')
    def duplicated(self, *,
            axis: int = 0,
            exclude_first: bool = False,
            exclude_last: bool = False) -> 'Batch':
        '''
        Return an axis-sized Boolean :obj:`Series` that shows True for all rows (axis 0) or columns (axis 1) duplicated.

        Args:
            {axis}
            {exclude_first}
            {exclude_last}
        '''
        return self._apply_attr(
                attr='duplicated',
                axis=axis,
                exclude_first=exclude_first,
                exclude_last=exclude_last,
                )

    @doc_inject(selector='duplicated')
    def drop_duplicated(self, *,
            axis: int = 0,
            exclude_first: bool = False,
            exclude_last: bool = False
            ) -> 'Batch':
        '''
        Return a :obj:`Batch` with contained :obj:`Frame` with duplicated rows (axis 0) or columns (axis 1) removed. All values in the row or column are compared to determine duplication.

        Args:
            {axis}
            {exclude_first}
            {exclude_last}
        '''
        return self._apply_attr(
                attr='drop_duplicated',
                axis=axis,
                exclude_first=exclude_first,
                exclude_last=exclude_last,
                )

    # as only useful on Frame, perhaps skip?
    # def set_index(self,
    # def set_index_hierarchy(self,
    # def unset_index(self, *,

    def __round__(self, decimals: int = 0) -> 'Batch':
        '''
        Return a :obj:`Batch` with contained :obj:`Frame` rounded to the given decimals. Negative decimals round to the left of the decimal point.

        Args:
            decimals: number of decimals to round to.
        '''
        return self._apply_attr(
                attr='__round__',
                decimals=decimals,
                )

    def roll(self,
            index: int = 0,
            columns: int = 0,
            include_index: bool = False,
            include_columns: bool = False,
            ) -> 'Batch':
        '''
        Roll columns and/or rows by positive or negative integer counts, where columns and/or rows roll around the axis.

        Args:
            include_index: Determine if index is included in index-wise rotation.
            include_columns: Determine if column index is included in index-wise rotation.
        '''
        return self._apply_attr(
                attr='roll',
                index=index,
                columns=columns,
                include_index=include_index,
                include_columns=include_columns,
                )

    def shift(self,
            index: int = 0,
            columns: int = 0,
            fill_value: tp.Any = np.nan) -> 'Batch':
        '''
        Shift columns and/or rows by positive or negative integer counts, where columns and/or rows fall of the axis and introduce missing values, filled by `fill_value`.
        '''
        return self._apply_attr(
                attr='shift',
                index=index,
                columns=columns,
                fill_value=fill_value,
                )

    #---------------------------------------------------------------------------
    # transformations resulting in changed dimensionality

    def count(self, *,
            skipna: bool = True,
            axis: int = 0,
            ) -> 'Batch':
        '''Apply count on contained Frames.
        '''
        return self._apply_attr(
                attr='count',
                skipna=skipna,
                axis=axis,
                )

    @doc_inject(selector='sample')
    def sample(self,
            index: tp.Optional[int] = None,
            columns: tp.Optional[int] = None,
            *,
            seed: tp.Optional[int] = None,
            ) -> 'Batch':
        '''Apply sample on contained Frames.

        Args:
            {index}
            {columns}
            {seed}
        '''
        return self._apply_attr(
                attr='sample',
                index=index,
                columns=columns,
                seed=seed,
                )

    @doc_inject(selector='head', class_name='Batch')
    def head(self, count: int = 5) -> 'Batch':
        '''{doc}

        Args:
            {count}
        '''
        return self._apply_attr(
                attr='head',
                count=count,
                )

    @doc_inject(selector='tail', class_name='Batch')
    def tail(self, count: int = 5) -> 'Batch':
        '''{doc}

        Args:
            {count}
        '''
        return self._apply_attr(
                attr='tail',
                count=count,
                )

    @doc_inject(selector='argminmax')
    def loc_min(self, *,
            skipna: bool = True,
            axis: int = 0
            ) -> 'Batch':
        '''
        Return the labels corresponding to the minimum value found.

        Args:
            {skipna}
            {axis}
        '''
        return self._apply_attr(
                attr='loc_min',
                skipna=skipna,
                axis=axis,
                )

    @doc_inject(selector='argminmax')
    def iloc_min(self, *,
            skipna: bool = True,
            axis: int = 0
            ) -> 'Batch':
        '''
        Return the integer indices corresponding to the minimum values found.

        Args:
            {skipna}
            {axis}
        '''
        return self._apply_attr(
                attr='iloc_min',
                skipna=skipna,
                axis=axis,
                )

    @doc_inject(selector='argminmax')
    def loc_max(self, *,
            skipna: bool = True,
            axis: int = 0
            ) -> 'Batch':
        '''
        Return the labels corresponding to the maximum values found.

        Args:
            {skipna}
            {axis}
        '''
        return self._apply_attr(
                attr='loc_max',
                skipna=skipna,
                axis=axis,
                )

    @doc_inject(selector='argminmax')
    def iloc_max(self, *,
            skipna: bool = True,
            axis: int = 0
            ) -> 'Batch':
        '''
        Return the integer indices corresponding to the maximum values found.

        Args:
            {skipna}
            {axis}
        '''
        return self._apply_attr(
                attr='iloc_max',
                skipna=skipna,
                axis=axis,
                )

    def cov(self, *,
            axis: int = 1,
            ddof: int = 1,
            ) -> 'Batch':
        '''
        Compute a covariance matrix.

        Args:
            axis: if 0, each row represents a variable, with observations as columns; if 1, each column represents a variable, with observations as rows. Defaults to 1.
            ddof: Delta degrees of freedom, defaults to 1.
        '''
        return self._apply_attr(
                attr='cov',
                axis=axis,
                ddof=ddof,
                )


    #---------------------------------------------------------------------------
    # utility function to numpy array

    def unique(self, *,
            axis: tp.Optional[int] = None,
            ) -> 'Batch':
        '''
        Return a NumPy array of unqiue values. If the axis argument is provied, uniqueness is determined by columns or row.

        '''
        return self._apply_attr(
                attr='unique',
                axis=axis,
                )

    #---------------------------------------------------------------------------
    # exporter

    def to_frame(self, *,
            axis: int = 0,
            union: bool = True,
            index: tp.Optional[tp.Union[IndexInitializer, IndexAutoFactoryType]] = None,
            columns: tp.Optional[tp.Union[IndexInitializer, IndexAutoFactoryType]] = None,
            name: NameType = None,
            fill_value: object = np.nan,
            consolidate_blocks: bool = False
        ) -> Frame:
        '''
        Consolidate stored :obj:`Frame` into a new :obj:`Frame` using the stored labels as the index on the provided ``axis`` using :obj:`Frame.from_concat`. This assumes that that the contained :obj:`Frame` have been reduced to single dimension along the provided `axis`.
        '''
        labels = []
        containers: tp.List[FrameOrSeries] = []
        ndim1d = True
        for label, container in self._items:
            labels.append(label)
            ndim1d &= container.ndim == 1
            containers.append(container)

        name = name if name is not None else self._name

        if ndim1d:
            if axis == 0 and index is None:
                index = labels
            if axis == 1 and columns is None:
                columns = labels
            # import ipdb; ipdb.set_trace()
            return Frame.from_concat( #type: ignore
                    containers,
                    axis=axis,
                    union=union,
                    index=index,
                    columns=columns,
                    name=name,
                    fill_value=fill_value,
                    consolidate_blocks=consolidate_blocks,
                    )
        # produce a hierarchical index to return all Frames
        f = Frame.from_concat_items(
                zip(labels, containers),
                axis=axis,
                union=union,
                name=name,
                fill_value=fill_value,
                consolidate_blocks=consolidate_blocks,
                )
        if index is not None or columns is not None:
            # this relabels, as that is how Frame.from_concat works
            f = f.relabel(index=index, columns=columns)
        return f

    def to_bus(self) -> 'Bus':
        '''Realize the :obj:`Batch` as an :obj:`Bus`. Note that, as a :obj:`Bus` must have all labels (even if :obj:`Frame` are loaded lazily), this :obj:`Batch` will be exhausted.
        '''
        series = Series.from_items(
                self.items(),
                name=self._name,
                dtype=DTYPE_OBJECT)

        return Bus(series, config=self._config)


